"""Low-level intrinsics to expose external functions
"""

# This code has heavily inspired in the numba.extending.intrisic code
from typing import Dict, List, Union
from collections import defaultdict
from numba.core import extending, funcdesc, types, typing
from rbc.typesystem import Type
from rbc.targetinfo import TargetInfo


class External:
    @classmethod
    def external(cls, *args):
        """
        Parameters
        ----------
        signature : object (str, ctypes function, python callable, numba function)
            Any object convertible to a Numba function via Type.fromobject(...).tonumba()
        """
        ts = defaultdict(list)
        key = None
        for signature in args:
            with TargetInfo.dummy():
                t = Type.fromobject(signature)
            if not t.is_function:
                raise ValueError("signature must represent a function type")

            if not t.name:
                raise ValueError(
                    f"external function name not specified for signature {signature}"
                )

            if key is None:
                key = t.name
            if not key:
                raise ValueError(
                    f"external function name not specified for signature {signature}"
                )

            for device in [a for a in t.annotation() or [] if a in ["CPU", "GPU"]] or [
                "CPU",
                "GPU",
            ]:
                ts[device].append(signature)

        obj = cls(key, ts)
        obj.register()
        return obj

    def __init__(
        self,
        key: str,
        signatures: Dict[str, List[Union[str, types.FunctionType, Type]]],
    ):
        """
        Parameters
        ----------
        key : str
            The key of the external function for typing
        signatures : List of function signatures
            A list of function type signature. i.e. 'int64 fn(int64, float64)'
        """
        self._signatures = signatures
        self.key = key

    def __str__(self):
        a = []
        for device in self._signatures:
            for t in self._signatures[device]:
                a.append(str(t))
        return ", ".join(a)

    @property
    def name(self):
        return self.key

    def match_signature(self, atypes):
        # Code here is the same found in remotejit.py::Signature::best_match
        device = "CPU" if TargetInfo().is_cpu else "GPU"
        available_types = tuple(map(Type.fromobject, self._signatures[device]))
        ftype = None
        match_penalty = None
        for typ in available_types:
            penalty = typ.match(atypes)
            if penalty is not None:
                if ftype is None or penalty < match_penalty:
                    ftype = typ
                    match_penalty = penalty
        if ftype is None:
            satypes = ", ".join(map(str, atypes))
            available = "; ".join(map(str, available_types))
            raise TypeError(
                f"found no matching function type to given argument types"
                f" `{satypes}` and device `{device}`. Available function types: {available}"
            )
        return ftype

    def get_codegen(self):
        # lowering
        def codegen(context, builder, sig, args):
            # Need to retrieve the function name again
            atypes = tuple(map(Type.fromobject, sig.args))
            t = self.match_signature(atypes)
            fn_name = t.name
            fndesc = funcdesc.ExternalFunctionDescriptor(
                fn_name, sig.return_type, sig.args
            )
            func = context.declare_external_function(builder.module, fndesc)
            return builder.call(func, args)

        return codegen

    def register(self):
        # typing
        class ExternalTemplate(typing.templates.AbstractTemplate):
            obj = self
            key = self.key

            def generic(self, args, kws):
                # get the correct signature and function name for the current device
                atypes = tuple(map(Type.fromobject, args))
                t = self.obj.match_signature(atypes)
                TargetInfo().add_external(t.name)

                codegen = self.obj.get_codegen()
                extending.lower_builtin(self.key, *t.tonumba().args)(codegen)
                return t.tonumba()

        typing.templates.infer(ExternalTemplate)
        typing.templates.infer_global(self, types.Function(ExternalTemplate))

    def __call__(self, *args, **kwargs):
        """
        This is only defined to pretend to be a callable from CPython.
        """
        msg = f"{self.name} is not usable in pure-python"
        raise NotImplementedError(msg)


external = External.external



'''Omnisci Bytes type that corresponds to Omnisci type TEXT ENCODED NONE.
'''

__all__ = ['OmnisciTextEncodingDictType', 'TextEncodingDict']

from .omnisci_metatype import OmnisciMetaType
from rbc import typesystem


class OmnisciTextEncodingDictType(typesystem.Type):
    """Omnisci Text Encoding Dict type for RBC typesystem.
    """

    @property
    def __typesystem_type__(self):
        return typesystem.Type('int32')


class TextEncodingDict(object, metaclass=OmnisciMetaType):
    pass



"""https://en.cppreference.com/w/c/numeric/math
"""


from collections import namedtuple
from . import make_intrinsic

_arg = namedtuple("arg", ("name", "ty"))

cmath = {
    # Trigonometric
    "cos": ("double", [_arg(name="x", ty="double")]),
    "cosf": ("float", [_arg(name="x", ty="float")]),
    "sin": ("double", [_arg(name="x", ty="double")]),
    "sinf": ("float", [_arg(name="x", ty="float")]),
    "tan": ("double", [_arg(name="x", ty="double")]),
    "tanf": ("float", [_arg(name="x", ty="float")]),
    "acos": ("double", [_arg(name="x", ty="double")]),
    "acosf": ("float", [_arg(name="x", ty="float")]),
    "asin": ("double", [_arg(name="x", ty="double")]),
    "asinf": ("float", [_arg(name="x", ty="float")]),
    "atan": ("double", [_arg(name="x", ty="double")]),
    "atanf": ("float", [_arg(name="x", ty="float")]),
    "atan2": ("double", [_arg(name="y", ty="double"), _arg(name="x", ty="double")]),
    "atan2f": ("float", [_arg(name="y", ty="float"), _arg(name="x", ty="float")]),
    # Hyperbolic
    "cosh": ("double", [_arg(name="x", ty="double")]),
    "coshf": ("float", [_arg(name="x", ty="float")]),
    "sinh": ("double", [_arg(name="x", ty="double")]),
    "sinhf": ("float", [_arg(name="x", ty="float")]),
    "tanh": ("double", [_arg(name="x", ty="double")]),
    "tanhf": ("float", [_arg(name="x", ty="float")]),
    "acosh": ("double", [_arg(name="x", ty="double")]),
    "acoshf": ("float", [_arg(name="x", ty="float")]),
    "asinh": ("double", [_arg(name="x", ty="double")]),
    "asinhf": ("float", [_arg(name="x", ty="float")]),
    "atanh": ("double", [_arg(name="x", ty="double")]),
    "atanhf": ("float", [_arg(name="x", ty="float")]),
    # Exponential and logarithmic functions
    "exp": ("double", [_arg(name="x", ty="double")]),
    "expf": ("float", [_arg(name="x", ty="float")]),
    "frexp": ("double", [_arg(name="x", ty="double"), _arg(name="exp", ty="int*")]),
    "frexpf": ("float", [_arg(name="x", ty="float"), _arg(name="exp", ty="int*")]),
    "ldexp": ("double", [_arg(name="x", ty="double"), _arg(name="exp", ty="int")]),
    "ldexpf": ("float", [_arg(name="x", ty="float"), _arg(name="exp", ty="int")]),
    "log": ("double", [_arg(name="x", ty="double")]),
    "logf": ("float", [_arg(name="x", ty="float")]),
    "log10": ("double", [_arg(name="x", ty="double")]),
    "log10f": ("float", [_arg(name="x", ty="float")]),
    "modf": ("double", [_arg(name="x", ty="double"), _arg(name="intpart", ty="double*")]),
    "modff": ("float", [_arg(name="x", ty="float"), _arg(name="intpart", ty="float*")]),
    "exp2": ("double", [_arg(name="x", ty="double")]),
    "exp2f": ("float", [_arg(name="x", ty="float")]),
    "expm1": ("double", [_arg(name="x", ty="double")]),
    "expm1f": ("float", [_arg(name="x", ty="float")]),
    "ilogb": ("double", [_arg(name="x", ty="double")]),
    "ilogbf": ("float", [_arg(name="x", ty="float")]),
    "log1p": ("double", [_arg(name="x", ty="double")]),
    "log1pf": ("float", [_arg(name="x", ty="float")]),
    "log2": ("double", [_arg(name="x", ty="double")]),
    "log2f": ("float", [_arg(name="x", ty="float")]),
    "logb": ("double", [_arg(name="x", ty="double")]),
    "logbf": ("float", [_arg(name="x", ty="float")]),
    # power functions
    "pow": (
        "double",
        [_arg(name="base", ty="double"), _arg(name="exponent", ty="double")],
    ),
    "powf": (
        "float",
        [_arg(name="base", ty="float"), _arg(name="exponent", ty="float")],
    ),
    "sqrt": ("double", [_arg(name="x", ty="double")]),
    "sqrtf": ("float", [_arg(name="x", ty="float")]),
    "cbrt": ("double", [_arg(name="x", ty="double")]),
    "cbrtf": ("float", [_arg(name="x", ty="float")]),
    "hypot": ("double", [_arg(name="x", ty="double"), _arg(name="y", ty="double")]),
    "hypotf": ("float", [_arg(name="x", ty="float"), _arg(name="y", ty="float")]),
    # error and gamma functions
    "erf": ("double", [_arg(name="x", ty="double")]),
    "erff": ("float", [_arg(name="x", ty="float")]),
    "erfc": ("double", [_arg(name="x", ty="double")]),
    "erfcf": ("float", [_arg(name="x", ty="float")]),
    "tgamma": ("double", [_arg(name="x", ty="double")]),
    "tgammaf": ("float", [_arg(name="x", ty="float")]),
    "lgamma": ("double", [_arg(name="x", ty="double")]),
    "lgammaf": ("float", [_arg(name="x", ty="float")]),
    # Rounding
    "ceil": ("double", [_arg(name="x", ty="double")]),
    "ceilf": ("float", [_arg(name="x", ty="float")]),
    "floor": ("double", [_arg(name="x", ty="double")]),
    "floorf": ("float", [_arg(name="x", ty="float")]),
    "fmod": (
        "double",
        [_arg(name="numer", ty="double"), _arg(name="denom", ty="double")],
    ),
    "fmodf": (
        "float",
        [_arg(name="numer", ty="float"), _arg(name="denom", ty="float")],
    ),
    "trunc": ("double", [_arg(name="x", ty="double")]),
    "truncf": ("float", [_arg(name="x", ty="float")]),
    "round": ("double", [_arg(name="x", ty="double")]),
    "roundf": ("float", [_arg(name="x", ty="float")]),
    "lround": ("long int", [_arg(name="x", ty="double")]),
    "lroundf": ("long int", [_arg(name="x", ty="float")]),
    "llround": ("long long int", [_arg(name="x", ty="double")]),
    "llroundf": ("long long int", [_arg(name="x", ty="float")]),
    "rint": ("double", [_arg(name="x", ty="double")]),
    "rintf": ("float", [_arg(name="x", ty="float")]),
    "lrint": ("long int", [_arg(name="x", ty="double")]),
    "lrintf": ("long int", [_arg(name="x", ty="float")]),
    "llrint": ("long long int", [_arg(name="x", ty="double")]),
    "llrintf": ("long long int", [_arg(name="x", ty="float")]),
    "nearbyint": ("double", [_arg(name="x", ty="double")]),
    "nearbyintf": ("float", [_arg(name="x", ty="float")]),
    "remainder": (
        "double",
        [_arg(name="numer", ty="double"), _arg(name="denom", ty="double")],
    ),
    "remainderf": (
        "float",
        [_arg(name="numer", ty="float"), _arg(name="denom", ty="float")],
    ),
    # Floating-point manipulation
    "copysign": (
        "double",
        [_arg(name="x", ty="double"), _arg(name="y", ty="double")],
    ),
    "copysignf": (
        "float",
        [_arg(name="x", ty="float"), _arg(name="y", ty="float")],
    ),
    "nan": ("double", [_arg(name="tagp", ty="const char*")]),
    "nanf": ("float", [_arg(name="tagp", ty="const char*")]),
    "nextafter": (
        "double",
        [_arg(name="x", ty="double"), _arg(name="y", ty="double")],
    ),
    "nextafterf": (
        "float",
        [_arg(name="x", ty="float"), _arg(name="y", ty="float")],
    ),
    "nexttoward": (
        "double",
        [_arg(name="x", ty="double"), _arg(name="y", ty="double")],
    ),
    "nexttowardf": (
        "float",
        [_arg(name="x", ty="float"), _arg(name="y", ty="float")],
    ),
    # Minimum, maximum, difference functions
    "fdim": ("double", [_arg(name="x", ty="double"), _arg(name="y", ty="double")]),
    "fdimf": ("float", [_arg(name="x", ty="float"), _arg(name="y", ty="float")]),
    "fmax": ("double", [_arg(name="x", ty="double"), _arg(name="y", ty="double")]),
    "fmaxf": ("float", [_arg(name="x", ty="float"), _arg(name="y", ty="float")]),
    "fmin": ("double", [_arg(name="x", ty="double"), _arg(name="y", ty="double")]),
    "fminf": ("float", [_arg(name="x", ty="float"), _arg(name="y", ty="float")]),
    # Other functions
    "fabs": ("double", [_arg(name="x", ty="double")]),
    "fabsf": ("float", [_arg(name="x", ty="float")]),
    "abs": ("long long int", [_arg(name="x", ty="double")]),
    "absf": ("long", [_arg(name="x", ty="float")]),
    "fma": (
        "double",
        [
            _arg(name="x", ty="double"),
            _arg(name="y", ty="double"),
            _arg(name="z", ty="double"),
        ],
    ),
    "fmaf": (
        "float",
        [
            _arg(name="x", ty="float"),
            _arg(name="y", ty="float"),
            _arg(name="z", ty="float"),
        ],
    ),
}


for fname, (retty, args) in cmath.items():
    argnames = [arg.name for arg in args]
    doc = f"C math function {fname}"
    fn = make_intrinsic(fname, retty, argnames, __name__, globals(), doc)



import pytest
from rbc.tests import omnisci_fixture


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), minimal_version=(5, 6)):
        define(o)

        def require_loadtime(kind, _cache=[None]):
            msg = _cache[0]
            if kind == 'lt':
                try:
                    if msg is not None:
                        raise msg
                    if o.has_cuda:
                        o.sql_execute('select lt_device_selection_udf_gpu(0)')
                    else:
                        o.sql_execute('select lt_device_selection_udf_cpu(0)')
                except Exception as msg:
                    _cache[0] = msg
                    pytest.skip(f'test requires load-time device selection UDFs ({msg}):'
                                f' run server with `--udf ../Tests/device_selection_samples.cpp`'
                                ' or check server logs for errors')

        o.require_loadtime = require_loadtime
        yield o


funcs = ('any', 'cpu', 'gpu', 'both')
kinds = ('rt', 'ct', 'lt')

cpu_device_code = 0x637075
gpu_device_code = 0x677075
any_device_code = 0x616e79
err_device_code = 0x657272


def decode(c):
    return {cpu_device_code: 'cpu',
            gpu_device_code: 'gpu',
            any_device_code: 'any',
            err_device_code: 'err'}[c]


def get_single_result(omnisci, kind, func):
    if func == 'any':
        if kind == 'rt':
            return any_device_code
        return gpu_device_code if omnisci.has_cuda else cpu_device_code
    if func == 'cpu':
        return cpu_device_code
    if func == 'gpu':
        return gpu_device_code if omnisci.has_cuda else err_device_code
    if func == 'both':
        return gpu_device_code if omnisci.has_cuda else cpu_device_code
    raise NotImplementedError(repr((kind, func)))


def get_pair_result(omnisci, ext, kind1, func1, kind2, func2,
                    allow_query_step_cpu_retry=False):
    r1 = get_single_result(omnisci, kind1, func1)
    r2 = get_single_result(omnisci, kind2, func2)
    if r1 == err_device_code or r2 == err_device_code:
        return err_device_code
    if allow_query_step_cpu_retry and ext in {'udtf/udf', 'udf/udtf'}:
        # with --allow-query-step-cpu-retry=1, UDTF and UDF execution
        # targets are independent
        return r1, r2
    if func1 == 'cpu' and func2 == 'gpu':
        return err_device_code
    if func1 == 'gpu' and func2 == 'cpu':
        return err_device_code
    if r1 == cpu_device_code or r2 == cpu_device_code:
        if any_device_code not in [r1, r2]:
            return cpu_device_code, cpu_device_code
    return r1, r2


def execute1(omnisci, query):
    try:
        _, result = omnisci.sql_execute(query)
    except Exception as msg:
        return msg
    else:
        result = list(result)
        return decode(result[0][0])


def execute2(omnisci, query):
    try:
        _, result = omnisci.sql_execute(query)
    except Exception as msg:
        return msg
    else:
        result = list(result)
        return decode(result[0][0]), decode(result[0][1])


def get_worker1(omnisci, ext, kind, func):
    expected = decode(get_single_result(omnisci, kind, func))
    if ext == 'udf':
        query = f'select {kind}_device_selection_{ext}_{func}(0)'
        return execute1, query, expected
    if ext == 'udtf':
        query = (f'select out0 from table({kind}_device_selection_{ext}_{func}'
                 f'(cursor(select i4 from {omnisci.table_name})))')
        return execute1, query, expected
    raise NotImplementedError(repr((ext, kind, func)))


def get_worker2(omnisci, ext, kind1, func1, kind2, func2,
                allow_query_step_cpu_retry=False):
    expected = get_pair_result(omnisci, ext, kind1, func1, kind2, func2,
                               allow_query_step_cpu_retry=allow_query_step_cpu_retry)
    if isinstance(expected, tuple):
        expected = tuple(map(decode, expected))
    else:
        expected = decode(expected)
    if ext == 'udf/udf':
        query = (f'select {kind1}_device_selection_udf_{func1}(0),'
                 f' {kind2}_device_selection_udf_{func2}(0)')
        return execute2, query, expected
    if ext == 'udtf/udf':
        query = (f'select out0 from table({kind1}_device_selection_udtf_{func1}'
                 f'(cursor(select {kind2}_device_selection_udf_{func2}(i4)'
                 f' from {omnisci.table_name})))')
        return execute1, query, expected[0] if isinstance(expected, tuple) else expected
    raise NotImplementedError(repr((ext, kind1, func1, kind2, func2)))


def define(omnisci):

    @omnisci('int32(int32)')
    def rt_device_selection_udf_any(x):
        # cannot determine which device is actually used
        return any_device_code

    @omnisci('int32(int32)', devices=['cpu'])
    def rt_device_selection_udf_cpu(x):
        return cpu_device_code

    @omnisci('int32(int32)', devices=['gpu'])
    def rt_device_selection_udf_gpu(x):
        return gpu_device_code

    @omnisci('int32(int32)', devices=['cpu'])  # NOQA
    def rt_device_selection_udf_both(x):  # NOQA
        return cpu_device_code

    @omnisci('int32(int32)', devices=['gpu'])  # NOQA
    def rt_device_selection_udf_both(x):  # NOQA
        return gpu_device_code

    @omnisci('int32(Column<int32>, OutputColumn<int32>)')
    def rt_device_selection_udtf_any(x, out):
        # cannot determine which device is actually used
        out[0] = any_device_code
        return 1

    @omnisci('int32(Column<int32>, OutputColumn<int32>)', devices=['cpu'])
    def rt_device_selection_udtf_cpu(x, out):
        out[0] = cpu_device_code
        return 1

    @omnisci('int32(Column<int32>, OutputColumn<int32>)', devices=['gpu'])
    def rt_device_selection_udtf_gpu(x, out):
        out[0] = gpu_device_code
        return 1

    @omnisci('int32(Column<int32>, OutputColumn<int32>)', devices=['cpu'])  # NOQA
    def rt_device_selection_udtf_both(x, out):  # NOQA
        out[0] = cpu_device_code
        return 1

    @omnisci('int32(Column<int32>, OutputColumn<int32>)', devices=['gpu'])  # NOQA
    def rt_device_selection_udtf_both(x, out):  # NOQA
        out[0] = gpu_device_code
        return 1


@pytest.mark.parametrize("func", funcs)
@pytest.mark.parametrize("ext", ['udf', 'udtf'])
@pytest.mark.parametrize("kind", ['rt', 'ct', 'lt'])
def test_device_selection_single(omnisci, func, ext, kind):
    omnisci.require_loadtime(kind)

    if kind == 'lt' and ext == 'udtf':
        pytest.skip('Load-time UDTFs not supported')

    execute, query, expected = get_worker1(omnisci, ext, kind, func)
    result = execute(omnisci, query)
    if isinstance(result, Exception):
        assert expected == decode(err_device_code), str(result)
    else:
        assert expected == result


@pytest.mark.parametrize("func12", ['any/any', 'any/cpu', 'any/gpu', 'any/both',
                                    'both/cpu', 'both/gpu', 'cpu/gpu', 'cpu/cpu',
                                    'gpu/gpu', 'both/both'])
@pytest.mark.parametrize("ext", ['udf/udf', 'udtf/udf'])
@pytest.mark.parametrize("kind2", kinds)
@pytest.mark.parametrize("kind1", kinds)
def test_device_selection_pair(omnisci, func12, ext, kind2, kind1):
    func12 = tuple(func12.split('/'))
    func1, func2 = func12

    omnisci.require_loadtime(kind1)
    omnisci.require_loadtime(kind2)
    if 'udtf' in ext and kind1 == 'lt':
        pytest.skip('Load-time UDTFs not supported')

    # Omniscidb server option --allow-query-step-cpu-retry=1 was
    # introduced in omniscidb-internal PR 5798, here assuming the
    # default value.
    allow_query_step_cpu_retry = omnisci.version[:2] >= (5, 8)

    execute, query, expected = get_worker2(
        omnisci, ext, kind1, func1, kind2, func2,
        allow_query_step_cpu_retry=allow_query_step_cpu_retry)
    result = execute(omnisci, query)
    if isinstance(result, Exception):
        assert expected == decode(err_device_code), str(result)
    else:
        assert expected == result
